/* ============================================================================ */
/*		SD_card.h : SD Card Access Header File				*/
/* ============================================================================ */
/*      Designed and programmed by Duck-Yong Yoon in 2009.  */



#define BV(bit)		(1<<(bit))		/* bit processing */
#define cbi(reg,bit)	reg &= ~(BV(bit))
#define sbi(reg,bit)	reg |= (BV(bit))







void Initialize_SPI(void);
void SPI_Mode0_250kHz(void);			// SPI mode 0, 250kHz
void SPI_Mode0_8MHz(void);			    // SPI mode 0, 8MHz
unsigned char SPI_char(U08 data);		// send a byte to SPI

void Initialize_SD(void);			// initialize SD card
void SD_command(U08 command, U32 sector);	// send SD card command
void SD_read_sector(U32 sector, U08* buffer);	// read a sector of SD card
    
/* ---------------------------------------------------------------------------- */

#define CS_SD		PB0			// SD card chip select signal

#define CMD0		0			// GO_IDLE_STATE
#define CMD8		8			// SEND_IF_COND

#define CMD17		17			// READ_SINGLE_BLOCK
#define CMD55		55			// APP_CMD
#define CMD58		58			// READ_OCR
#define ACMD41		41			// SD_SEND_OP_COND

#define VER1X		0			// Ver1.X SD Memory Card
#define VER2X		1			// Ver2.X SD Memory Card
#define VER2X_SC	0			// Ver2.X Standard Capacity SD Memory Card
#define VER2X_HC	1			// Ver2.X High Capacity SD Memory Card(SDHC)

unsigned char SD_type;				// SD card version(Ver1.X or Ver2.0)

/* ---------------------------------------------------------------------------- */

void Initialize_SPI(void)			/* initialize SPI */
{
  DDRB |= 0x07;					//0x16    PB0, 2, 1 = output
  PORTB |= 0x03;				// CS_SD = SCK = high
}

void SPI_Mode0_250kHz(void)			/* SPI mode 0, 250kHz */
{
  SPCR = 0x53;					// 16MHz/64 = 250kHz
  SPSR = 0x01;
}

void SPI_Mode0_8MHz(void)			/* SPI mode 0, 8MHz */
{
  SPCR = 0x50;					// 16MHz/2 = 8MHz
  SPSR = 0x01;
}

unsigned char SPI_char(U08 data)		/* send a byte to SPI */
{
  SPDR = data;
  while(!(SPSR & 0x80));

  return SPDR;
}

/* ---------------------------------------------------------------------------- */

void Initialize_SD(void)			/* initialize SD/SDHC card */
{
  unsigned char i, j;
  unsigned char R1, CMD_flag, R3[5], R7[5];

  sbi(DDRB, CS_SD);				// -CS_SD is output

  TFT_string(0,0,Green,  Black,"******************************");
  TFT_string(0,2,Magenta,Black,"     SD/SDHC ī  ʱȭ     ");
  TFT_string(0,4,Green,  Black,"******************************");

	delay_ms(100);
  SPI_Mode0_250kHz();				//  must use below 400kHz
  sbi(PORTB, CS_SD);				// -CS_SD = 1
  for(i = 1; i <= 10; i++) SPI_char(0xFF);	// set SPI mode(80 clocks)

  cbi(PORTB, CS_SD);				// -CS_SD = 0

  TFT_string(0,8,Magenta,Black,"CMD0 : ");
  TFT_color(Cyan, Black);
  SD_command(CMD0, 0);				// send CMD0(reset and go to SPI mode)
  CMD_flag = 0;
  for(i = 0; i < 5; i++)			// display R1
    { R1 = SPI_char(0xFF);
      TFT_hexadecimal(R1,2);
      TFT_English(' ');
      if(R1 == 0x01)
        CMD_flag = 1;
    }
  if(CMD_flag == 1)
    TFT_string(22,8,Green,  Black,"/ OK !");
  else
    { TFT_string(22,8,Red,    Black,"/ Error!");
      TFT_string(5,22,Magenta,Black,"SD ī尡 ϴ !");

      while(1);
    }

  sbi(PORTB, CS_SD);				// -CS_SD = 1

  SPI_Mode0_8MHz();				//  must use below 25MHz
  cbi(PORTB, CS_SD);				// -CS_SD = 0

  TFT_string(0,10,Magenta,Black,"CMD8 : ");
  TFT_color(Cyan, Black);
  SD_command(CMD8, 0x000001AA);			// send CMD8(check version)
  CMD_flag = 0;
  for(i = 0; i < 5; i++)
    { TFT_xy(7,10);
      R7[0] =  SPI_char(0xFF);
      TFT_hexadecimal(R7[0],2);
      TFT_English(' ');
      if(R7[0] == 0x05)				// if R1 = 0x05, Ver 1.X
        { SD_type = VER1X;
          CMD_flag = 1;
          TFT_string(22,10,Green,Black,"/ V1.X");
	  break;
	}
      else if(R7[0] == 0x01)			// if R1 = 0x01, Ver 2.X
        { for(j = 1; j <= 4; j++)
            { R7[j] = SPI_char(0xFF);
              TFT_hexadecimal(R7[j],2);
              TFT_English(' ');
            }
          if((R7[3] == 0x01) && (R7[4] == 0xAA))
	    { SD_type = VER2X;
	      CMD_flag = 1;
              TFT_string(22,10,Green,Black,"/ V2.X");
	      break;
            }
	}
    }
  SPI_char(0xFF);				// send dummy clocks
  if(CMD_flag == 0)
    { TFT_string(22,10,Red,    Black,"/ Error!");
      TFT_string( 0,22,Magenta,Black,"SD card with unknown format !");
 //     Beep_3times();
      while(1);
    }

  CMD_flag = 0;
  for(i = 0; i < 5; i++)			// send ACMD41(start initialization)
    { TFT_string(0,14,Magenta,Black,"CMD55: ");
      TFT_color(Cyan, Black);
      SD_command(CMD55, 0);
      for(j = 0; j < 5; j++)			// receive R1 = 0x01
        { R1 = SPI_char(0xFF);
          TFT_hexadecimal(R1,2);
          TFT_English(' ');
        }
      TFT_string(0,16,Magenta,Black,"ACMD41:");
      TFT_color(Cyan, Black);
      SD_command(ACMD41, 0x40000000);		// HCS = 1
      for(j = 0; j < 5; j++)			// receive R1 = 0x00
        { R1 = SPI_char(0xFF);
          TFT_hexadecimal(R1,2);
          TFT_English(' ');
          if(R1 == 0x00)
            CMD_flag = 1;
        }
    }
  if(CMD_flag == 1)
    TFT_string(22,16,Green,  Black,"/ OK !");
  else
    { TFT_string(22,16,Red,    Black,"/ Error!");
      TFT_string( 4,22,Magenta,Black,"SD card is not ready !");
 
      while(1);
    }

  if(SD_type == VER2X)
    { TFT_string(0,18,Magenta,Black,"CMD58: "); // send CMD58(check SDHC)
      TFT_color(Cyan, Black);
      SD_command(CMD58, 0);
      for(i = 0; i < 5; i++)
        { TFT_xy(7,18);
          R3[0] =  SPI_char(0xFF);
          TFT_hexadecimal(R3[0],2);
          TFT_English(' ');
          if(R3[0] == 0x00)
            { for(j = 1; j <= 4; j++)
                { R3[j] = SPI_char(0xFF);
                  TFT_hexadecimal(R3[j],2);
                  TFT_English(' ');
	        }
              break;
	    }
        }
      if((R3[1] & 0xC0) == 0xC0)		// if CCS = 1, High Capacity(SDHC)
        { SD_type = VER2X_HC;
          TFT_string(22,18,Magenta,Black,"(SDHC)");
        }
      else					// if CCS = 0, Standard Capacity
        SD_type = VER2X_SC;
      SPI_char(0xFF);				// send dummy clocks
    }

  sbi(PORTB, CS_SD);				// -CS_SD = 1

  TFT_string(6,19,Green,Black,"ʱȭ Ϸ !");
}

void SD_command(U08 command, U32 sector)	/* send SD card command */
{
  unsigned char CRC;

  SPI_char(command | 0x40);			// send command index
  SPI_char(sector >> 24);			// send 32-bit argument
  SPI_char(sector >> 16);
  SPI_char(sector >> 8);
  SPI_char(sector);


  if(command == CMD0)      CRC = 0x95;		// CRC for CMD0
  else if(command == CMD8) CRC = 0x87;		// CRC for CMD8(0x000001AA)
  else                     CRC = 0xFF;		// no CRC for other CMD

  SPI_char(CRC);				// send CRC
}

void SD_read_sector(U32 sector, U08 *buffer)	/* read a sector of SD card */
{
  unsigned int i;

  if((SD_type == VER1X) || (SD_type == VER2X_SC)) // if not SDHC, sector = sector*512
    sector <<= 9;

  cbi(PORTB, CS_SD);				// -CS_SD = 0

  SD_command(CMD17, sector);			// send CMD17(read a block)

  while(SPI_char(0xFF) != 0x00);		// wait for R1 = 0x00
  while(SPI_char(0xFF) != 0xFE);		// wait for Start Block Token = 0xFE

  for(i = 0; i < 512; i++)			// receive 512-byte data
    buffer[i] = SPI_char(0xFF);

  SPI_char(0xFF);				// receive CRC(2 bytes)
  SPI_char(0xFF);
  SPI_char(0xFF);				// send dummy clocks

  sbi(PORTB, CS_SD);				// -CS_SD = 1
}

