//**********************************************************
// **** MAIN routine FOR Interfacing microSD/SD CARD ******
//**********************************************************
//Controller: ATmega128(16 Mhz internal)
//Compiler	: AVR-GCC
//********************************************************

//Link to the Post: http://www.dharmanitech.com/2009/01/sd-card-interfacing-with-atmega8-fat32.html

#define F_CPU 16000000UL		//freq 8 MHz
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include "UART_routines.h"

#include "tftlcd.h"

//#include "image.h"
#include "touch.h"


#include "SD_card.h"
#include "SD_FAT32.h"


void TFT_filenumber(void);			// display file number
void TFT_filename(U08 file_number);		// display BMP file name
void TFT_display_BMP(U32 sector);		// display BMP file of SD card


volatile unsigned int timer = 0;		// 1 sec counter

unsigned char total_file;			// total file number
unsigned char file_number = 0;			// current file number
unsigned char BMP_type;				// BMP image type(P=portrait, L=landscape)




unsigned char line_width = 1;			// line width (default = middle)
unsigned int  user_color = White;		// user color (default = White)

void Draw_select(void)				/* draw selected color and line width */
{
  Block(10,267, 45,287, Black, Black);		// clear select box

  Rectangle(10,267, 45,287, user_color);	// line_width = 0 ?
  if(line_width >= 1)				// line_width = 1 ?
    { Rectangle(11,268, 44,286, user_color);
      Rectangle(12,269, 43,285, user_color);
    }
  if(line_width >= 2)				// line_width = 2 ?
    { Rectangle(13,270, 42,284, user_color);
      Rectangle(14,271, 41,283, user_color);
    }

  delay_ms(100);
}



void Title_screen(void)				/* draw title screen */
{
  Rectangle(0,0, 239,319, Green);		// draw outline of sheet
  Rectangle(1,1, 238,318, Green);
  Rectangle(2,2, 237,317, Green);

  TFT_string(7,1,White,Magenta,"׸ (TSC2003)");	// title

  Draw_select();				// select box

  TFT_string(2,37,Green,Black," C ");		// clear selection button
  Rectangle(10,292, 45,312, Green);
  Rectangle(11,293, 44,311, Green);

  Block(160,267, 180,287, White, White);	// line_width selection button
  Line(165,277, 175,277, Black);
  Block(185,267, 205,287, White, White);
  Line(190,276, 200,276, Black);
  Line(190,277, 200,277, Black);
  Line(190,278, 200,278, Black);
  Block(210,267, 230,287, White, White);
  Line(215,275, 225,275, Black);
  Line(215,276, 225,276, Black);
  Line(215,277, 225,277, Black);
  Line(215,278, 225,278, Black);
  Line(215,279, 225,279, Black);

  Block( 60,292,  80,312, White, White);	// color selection button
  Block( 85,292, 105,312, White, Red);
  Block(110,292, 130,312, White, Yellow);
  Block(135,292, 155,312, White, Green);
  Block(160,292, 180,312, White, Cyan);
  Block(185,292, 205,312, White, Blue);
  Block(210,292, 230,312, White, Magenta);
}




//*************************** MAIN *******************************//
int main(void)
{
//unsigned char option, error, data, FAT32_active;
//

//unsigned int i, x, y, color;
   unsigned int xo = 0, yo = 0; 
  unsigned char i; 
 
  unsigned char display_flag = 1;		// new screen display flag
 // unsigned char filename_flag = 1;		// file name display flag

  unsigned long BMP_start_sector[MAX_FILE];	// BMP file start sector

 
 init_devices();
delay_ms(500);
 //PORTD |= 0x04; //switching ON the LED (for testing purpose only)
 //C2003_Setup();
 TFT_SET();
Initialize_SPI();				// initialize SPI
 Initialize_SD();				// initialize SD card
Initialize_FAT32();	

delay_ms(700);delay_ms(700);
Initialize_TSC2003();
			// initialize FAT32 file system

 

 TFT_color_screen(MAGENTA);
 TX_NEWLINE;
 TX_NEWLINE;
 transmitString_F (PSTR("****************************************************"));
 TX_NEWLINE;
 transmitString_F (PSTR("   Signaltech microSD Card Testing.  "));
 TX_NEWLINE;
 transmitString_F (PSTR("****************************************************"));
 TX_NEWLINE;
  
  
  TFT_string(0, 6,Green,  Black,"******************************");
  TFT_string(0, 8,White,Magenta,"      BMP   ھٹ      ");
  TFT_string(0,10,Green,  Black,"******************************");

  delay_ms(3000);

 /*                              // clear all interrupt flags
TCCR1A = 0x00;                                // CTC mode(4), don't output OC1A
  TCCR1B = 0x0C;                                // 16MHz/256/(1+62499) = 1Hz
  TCCR1C = 0x00;
  OCR1A = 62499;
  TCNT1 = 0x0000;

  TIMSK |= 0x01;                                // enable OC1A interrupt
  TIFR = 0x2F; 
sei();
*/
  total_file = fatGetDirEntry(FirstDirCluster);	// get total file number
  for(i = 0; i < total_file; i++)		// get start address of all files
    BMP_start_sector[i] = fatClustToSect(file_start_cluster[i]);
 file_number = 0;


while(1)
{


if(display_flag == 1)	// display BMP file, file number, and file name
        { Get_long_filename(file_number);	//  check file name 
	  if(extension != 0x00424D50)		// if not BMP file, skip
            { 
			if(file_number < (total_file - 1))
	        file_number++;
	      else
	        file_number = 0;
	        }
          else					// if BMP file, display
	        {
			 TFT_display_BMP(BMP_start_sector[file_number]);
              TFT_filenumber();
	          TFT_filename(file_number);
	          display_flag = 0;
  	        }
          }
 Touch_input_TSC2003();			// input touch screen

 TFT_string(2,26,Green,Black,"Signaltech");

      if((x_touch != 0) && (y_touch != 0))
	{
	  if(file_number < (total_file - 1))	// if touch screen input, go next file
	    file_number++;
	  else
	    file_number = 0;

	  display_flag = 1;
	}
	
  } //While
}  //main  

void TFT_filenumber(void)			/* display file number */
{
  TFT_screen_mode(BMP_type);			// screen mode(Portrait or Landscape)

  TFT_xy(0,0);					// start position and color
  TFT_color(White, Transparent);

  TFT_English('(');
  TFT_unsigned_decimal(file_number+1,1,2);	// current file number
  TFT_English('/');
  TFT_unsigned_decimal(total_file,1,2);		// total file number
  TFT_English(')');
}


void TFT_filename(U08 file_number)		/* display BMP file name */
{

  unsigned char file_flag;

   file_flag = Get_long_filename(file_number);	// check file name

  if(file_flag == 0)				// short file name(8.3 format)
   TFT_short_filename(8,0,Yellow,Transparent);
  else if(file_flag == 1)			// long file name
   TFT_long_filename(8,0,Yellow,Transparent);
  else if(file_flag == 2)			// file name is longer than 195 characters
    TFT_string(8,0,White,Transparent,"* ϸ  ʰ *");
  else						// file name error
    TFT_string(8,0,White,Transparent,"*** ϸ  ***");

}




/* ---------------------------------------------------------------------------- */

struct BMP_header{				/* Header 14 bytes + InfoHeader 40 bytes */
 unsigned char signature[2];			// 'BM'
 unsigned long file_size;			// file size in bytes
 unsigned long reserved;			// unused (= 0)
 unsigned long data_offset;			// file offset to Raster Data

 unsigned long size;				// size of InfoHeader (= 40)
 unsigned long width;				// bitmap width
 unsigned long height;				// bitmap height
 unsigned int  planes;				// number of planes (= 1)
 unsigned int  bit_count;			// bits per pixel 
 unsigned long compression;			// type of compression   
 unsigned long image_size;			// size of image
 unsigned long x_pixels_per_m;			// horizontal resolution(pixels/meter)
 unsigned long y_pixels_per_m;			// vertical resolution(pixels/meter)
 unsigned long colors_used;			// number of actually used colors
 unsigned long colors_important;		// number of important colors. 0 = all
};

// Description : SD ī忡  BMP ̹  TFT-LCD ϱ
// sector = BMP ̹ ִ Ͻ ͹ȣ
//
//         [B]         [G]       [R]
//          5           6         5
//       11111000   11111100   11111000
//          &           &         &
//        buf[0]      buf[1]    buf[2]
//
//         >> 3        << 3      << 8 (BGR ¸ RGB · ٽ 迭)
//
//          [R]         [G]        [B]
// pixel = ?????      ??????      ????? = 16 bit

void TFT_display_BMP(U32 sector)		/* display BMP file of SD card */
{
  unsigned char buffer_control;
  unsigned int  i, pixel, sector_cnt;
  unsigned char BMPbuffer[512];   		// BMP data from SD card

  struct BMP_header *bmp_ptr;

  SD_read_sector(sector, BMPbuffer);		// read sector of BMP header

  bmp_ptr = (struct BMP_header *)BMPbuffer;

  if(bmp_ptr->width == 240)			// if 240*320, Portrait screen(bottom --> top)
    { BMP_type = 'P';  //p 
  LCD_WR_REG(0x0003,0x1000); // 8 x set GRAM write direction and BGR=1.
  LCD_WR_REG(0x0020,0x0000);  //  x=0
  LCD_WR_REG(0x0021,0x013f);  //  y=319
  LCD_WR_REG16(0x0022);    

    }
  else if(bmp_ptr->width == 320)		//     if 320*240, Landscape screen(left --> right)
    { BMP_type = 'L';  //pic ok
  LCD_WR_REG(0x0003,0x1008); // 0 x set GRAM write direction and BGR=1.
  LCD_WR_REG(0x0020,0x0000);  //  x=0
  LCD_WR_REG(0x0021,0x0000);  //  y=319
  LCD_WR_REG16(0x0022);  
    }

  for(i = 0x36; i <= 509; i += 3)
    { pixel  = (BMPbuffer[i] & 0xf8) >> 3;	// Blue
      pixel |= (BMPbuffer[i+1] & 0xfc) << 3;	// Green
      pixel |= (BMPbuffer[i+2] & 0xf8) << 8;	// Red
      LCD_WR_DATA16(pixel);
    }
  sector++;
  buffer_control = 1;

  for(sector_cnt = 0; sector_cnt <= 449; sector_cnt++)	// read next sector and output RGB data
    {
      switch (buffer_control)			// rotate case 1, 2, 3
	{ case 1: pixel  = (BMPbuffer[510] & 0xF8) >> 3;
		  pixel |= (BMPbuffer[511] & 0xFC) << 3;
		  SD_read_sector(sector, BMPbuffer);
		  pixel |= (BMPbuffer[0] & 0xF8) << 8;
		  LCD_WR_DATA16(pixel);

		  for(i = 1; i <= 509; i += 3)
		    { pixel  = (BMPbuffer[i] & 0xf8) >> 3;
		      pixel |= (BMPbuffer[i+1] & 0xfc) << 3;
		      pixel |= (BMPbuffer[i+2] & 0xf8) << 8;
		      LCD_WR_DATA16(pixel);
		    }
		  sector++;
		  buffer_control = 2;
		  break;

	  case 2: pixel  = (BMPbuffer[511] & 0xF8) >> 3;
		  SD_read_sector(sector, BMPbuffer);
		  pixel |= (BMPbuffer[0] & 0xFC) << 3;
		  pixel |= (BMPbuffer[1] & 0xF8) << 8;
		  LCD_WR_DATA16(pixel);

		  for(i = 2; i <= 509; i += 3)
		    { pixel  = (BMPbuffer[i] & 0xF8) >> 3;
		      pixel |= (BMPbuffer[i+1] & 0xFC) << 3;
		      pixel |= (BMPbuffer[i+2] & 0xF8) << 8;
		      LCD_WR_DATA16(pixel);
		    }
		   sector++;
		   buffer_control = 3;
		   break;

	  case 3: SD_read_sector(sector, BMPbuffer);
		  if(sector_cnt != 449)
		    { for(i = 0; i <= 509; i += 3)
			{ pixel  = (BMPbuffer[i] & 0xF8) >> 3;
			  pixel |= (BMPbuffer[i+1] & 0xFC) << 3;
			  pixel |= (BMPbuffer[i+2] & 0xF8) << 8;
			  LCD_WR_DATA16(pixel);
			}
		    }
		  else				// if last sector, read 54 bytes only
		    { for(i = 0; i < 54; i += 3)
			{ pixel  = (BMPbuffer[i] & 0xF8) >> 3;
			  pixel |= (BMPbuffer[i+1] & 0xFC) << 3;
			  pixel |= (BMPbuffer[i+2] & 0xF8) << 8;
			  LCD_WR_DATA16(pixel);
			}
		    }
		  sector++;
		  buffer_control = 1;
		  break;
	}
    }
  LCD_WR_REG(0x0003,0x1030); // set GRAM write direction and BGR=1.
}


void port_init(void)
{
  DDRA=0xff;
   PORTB = 0xFE;
   DDRB  = 0xF7;
////MISO line i/p, rest o/p
    DDRC=0xff;
    DDRD=0xfB;
	DDRE=0x7f;
	DDRF=0xff;

	PORTD= 0xff;
	PORTE= 0xff; 
}

//UART0 initialize
// desired baud rate: 19200
// actual: baud rate:19231 (0.2%)
// char size: 8 bit
// parity: Disabled
void uart0_init(void)
{

  UBRR0H = 0;                                   //8M => 19200, 16M => 38400 bps
  UBRR0L = 25;
  UCSR0A = 0x00;                                // asynchronous normal mode
  UCSR0B = 0x18;                                // interrupt enable, 8 data
  UCSR0C = 0x06;                                // no parity, 1 stop, 8 data

 /*UCSR0B = 0x00; //disable while setting baud rate
 UCSR0A = 0x00;
 UCSR0C = (1 << UMSEL0) | 0x06;
 UBRR0L = 0x19; //set baud rate lo
 UBRR0H = 0x00; //set baud rate hi
 UCSR0B = 0x18;
*/
}

//call this routine to initialize all peripherals
void init_devices(void)
{
 
 port_init();
 uart0_init();
// Timer/Counter0 (Overflow)
   
/*
  TCCR1A = 0x00;      // timer/counter1 regA CTC mode(4), don't output OC1A
  TCCR1B = 0x0C;      //  timer/counter1 regB  16MHz/256/(1+62499) = 1Hz
  TCCR1C = 0x00;  //timer/counter1 reg C
  OCR1A = 62499; //output compare regA1
  TCNT1 = 0x0000;  //timer/counter1

  TIMSK1 = 0x02;        //timer/counter1  enable OC1A interrupt
  TIFR1 = 0x2F;         // clear all interrupt flags
  */
 
}
